from maze_keeper import MazeObjects, MazeKeeper, ACTIONS
import random

class MazeGenerator:
    """
    When supplied with the size of the maze as NxM tuple (M - number of rows, M - number of columns), maze generator has
    to generate a maze of given size.
    Coordinates in the maze are such that c is the index of the column, r is the index of the row, indexing from top-left
    corner at (0,0).
    Maze is a 2-D list that consists of cells that are one of the elements of the MazeObjects enum, i.e.
    empty, obstacle, start or gold .
    Generated maze has to be feasible, i.e. there must exist valid path between the start and the gold.
    """

    def __init__(self, maze_size, step_limit):
        self.step_limit = step_limit
        self.maze_size = maze_size

    def generate_empty(self):
        """Creates an empty maze with start at top-left corner and gold at the opposite corner"""
        layout = []
        for r in range(self.maze_size[0]):
            row = []
            for c in range(self.maze_size[1]):
                row.append(MazeObjects.EMPTY)
            layout.append(row)
        layout[0][0] = MazeObjects.START
        layout[-1][-1] = MazeObjects.GOLD
        return layout

    def generate_random(self):
        """Creates random maze with origin and destination in opposing corners"""
        layout = []
        for r in range(self.maze_size[0]):
            row = []
            for c in range(self.maze_size[1]):
                row.append(random.choice([MazeObjects.EMPTY,MazeObjects.EMPTY,MazeObjects.EMPTY,MazeObjects.OBSTACLE]))
            layout.append(row)
        layout[0][0] = MazeObjects.START
        layout[-1][-1] = MazeObjects.GOLD
        return layout

    @staticmethod
    def is_feasible(layout):
        gold = MazeKeeper._get_unique_cell(layout, MazeObjects.GOLD)
        start = MazeKeeper._get_unique_cell(layout, MazeObjects.START)
        movable_positions = set(MazeKeeper._get_cell_positions(layout, [MazeObjects.EMPTY, MazeObjects.GOLD]))

        wave = {start}
        while wave:
            if gold in wave:
                return True
            expanding_cell = wave.pop()
            for action in ACTIONS.values():
                new_cell = MazeKeeper.add_tuples(expanding_cell, action)
                if new_cell in movable_positions:
                    movable_positions.remove(new_cell)
                    wave.add(new_cell)
        return False

    def generate(self):
        """
        Maze is built from cells that can be one of the elements of the MazeObjects enum,
        i.e. empty, obstacle, start or gold.
        :return: 2-D list, the layout of the maze

        ===YOUR MAZE GENERATION CODE GOES HERE===
        """
        while True:
            layout = self.generate_random()
            if self.is_feasible(layout):
                break
        return layout


if __name__ == '__main__':
    mg = MazeGenerator((5, 5), 100)
    print(mg.generate_random())
